@@ -49,7 +49,7 @@ sqlFetch(channel, "Atest")
 sqlDrop(channel, "Atest")
 close(channel)
 
-channel <- odbcConnect("testdb3", uid="ripley")
+channel <- odbcConnect("testdb3", uid="ripley", bulk_add="no")
 dates <- as.character(seq(as.Date("2004-01-01"), by="week", length=10))
 times <- paste(1:10, "05", "00", sep=":")
 Dtest <- data.frame(dates, times, logi=c(TRUE, NA, FALSE, FALSE, FALSE))
@@ -411,6 +411,11 @@ sqlDrop(channel, "hills test")
 sqlSave(channel, hills, "hills test", verbose=TRUE, fast=FALSE)
 sqlUpdate(channel, hills[11:15,], "hills test", verbose=TRUE, fast=FALSE)
 sqlDrop(channel, "hills test")
+
+A <- data.frame(a="Latin-1 accented chars:   < <")
+sqlDrop(channel, "A", errors = FALSE)
+sqlSave(channel, A, verbose=TRUE)
+sqlFetch(channel, "A")
 ## close the connection
 close(channel)
 
@@ -434,7 +439,7 @@ sqlUpdate(channel, foo, "USArrests", index = "state")
 sqlFetch(channel, "USArrests", rownames = "state", max = 5)
 sqlDrop(channel, "USArrests")
 
-qlDrop(channel, "hills test", errors = FALSE)
+sqlDrop(channel, "hills test", errors = FALSE)
 sqlSave(channel, hills, "hills test", verbose=TRUE)
 sqlUpdate(channel, hills[11:15,], "hills test", verbose=TRUE, fast=TRUE)
 sqlFetch(channel, "hills test")
